from pig_ui.constants import *
from pig_ui.ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.ui.ui_element import UIElement
from pig_ui.ui.ui_manager import UIM
class UIDropDown(UIElement):
    def __init__(self, app, pos, size, ux = None, draggable = False, **kwargs):
        kwargs['cb_lclick'] = self.toggle_dd
        kwargs['anchor'] = "tl"
        ux=UXWrapper(
            ux = [
                [UXRect(-1,Color(col),size=size),
                 UXText(text_get_callback=kwargs.get('title', ''))] for col in ('#484848', '#969696', '#ffffff', '#000000')
            ]
        )
        super().__init__(app, pos, size, ux, draggable, **kwargs)
        self.sub = []
        kwargs.get('ltext')
        kwargs.get('lcom')
    @property
    def text(self) -> str:
        return "test"
    def toggle_dd(self,*_):
        for uie in self.sub:
            uie.visible = not uie.visible
    def sub_callbacks(self,obj):
        i = self.sub.index(obj)
        self.sub_cbs[i](obj)
        for uie in self.sub:
            uie.visible = not uie.visible
    def set_subs(self, ltext: list[str], lcom: list[Callable] | None = None):
        for uie in self.sub:
            uie: UIElement
            uie.destroy()
        
        self.texts = []
        if lcom is None:
            lcom = [lambda *x: None for i in ltext]
        print(ltext, lcom)
        self.sub_cbs = lcom
        for i, (t, c) in enumerate(zip(ltext, lcom)):
            print(t,)
            self.texts.append(t)
            ux = [
                [UXRect(-1,Color(col),size=self.size),
                 UXText(text_get_callback=t)] for col in ('#484848', '#969696', '#ffffff', '#000000')
            ]
            
            uie = UIElement(
                self.app, 
                Vector2(0,(i+1) * self.size.y),
                self.size,
                draggable=False,
                ux=UXWrapper(ux), 
                parent=self,
                anchor="tl",
                cb_lclick=self.sub_callbacks,
                cb_dclick=lambda x: None,
                visible=False
                )
            print(self.abs_offset + Vector2(0,(i+1) * self.size.y), uie.abs_offset)
            self.sub.append(uie)


class UIManager:
    def __init__(self):
        self.QUEUE = []
        self.blocked = -1
        
    def add_to_queue(self,object):
        """
        This adds a ``UIElement`` to the render queue.
        """
        self.QUEUE.append(object)
        
    def remove_from_queue(self, object):
        """
        Removes the given ``UIElement`` from the QUEUE.
        """
        self.QUEUE.remove(object)
    
    def stick_on_top(self,object):
        """
        Takes the selected Object, if its exists in QUEUE and
        remove it from its old layer and re:adds it on the end of the set 
        """
        if object.uid in self.uids:
            i = self.QUEUE.index(object)
            self.QUEUE.remove(i)
            self.QUEUE.append(i)
    
    def __get_ordered_by_layers(self) -> tuple[list[int], list[int]]:
        """
        HUD LAYER = 1
        BG LAYER = 2
        """
        _q0 = [(o.layer + o.o_layer, o.uid) for o in self.QUEUE if not o.o_layer]
        _q1 = [(o.layer + o.o_layer, o.uid) for o in self.QUEUE if o.o_layer]
            
        _s0 = sorted(
            _q0,
            key = lambda x: x[0]
        )
        _s1 = sorted(
            _q1,
            key = lambda x: x[0]
        )
        
        return [i[1] for i in _s0], [i[1] for i in _s1]
    
    def __get_object_by_id(self, id: int):
        if id in self.uids:
            for o in self.QUEUE:
                if o.uid == id:
                    return o
        assert id not in self.uids
    
    @staticmethod
    def get_skip_unwanted_groups(object, groups) -> bool:
        """
        Skip unwanted groups & not visible uie's
        """
        return not object.visible or (groups and not object.group in groups)
    
    @staticmethod
    def get_blocked_state(object,blocked) -> bool:
        return blocked != -1 and blocked != object.uid
    
    @staticmethod
    def get_dead_uie(object, blocked) -> bool:
        """
        We encountered a softlock after:
            Pressing a button that makes itself invisbile
        This is a fix/workaround for this
        """
        return not object.visible and object.uid == blocked
    
    @staticmethod
    def get_key_reset(object) -> bool:
        return hasattr(object, 'pressed_keys') and object.pressed_keys and not object.event.KEYS
    
    def update(self, groups: list | tuple | None = None):
        ids = self.__get_ordered_by_layers()
        ids = ids[0] + ids[1]
        
        for id in ids[::-1]:
            object = self.__get_object_by_id(id)
            
            if UIManager.get_dead_uie(object, self.blocked):
                self.blocked = -1
                break
            
            if UIManager.get_skip_unwanted_groups(object,groups):
                continue
            
            if UIManager.get_blocked_state(object, self.blocked): 
                if hasattr(object,'reset'):
                    object.reset()
                continue # other UIE has higher priority because it is actually running(drag&drop)
            
            if UIManager.get_key_reset(object):
                object.pressed_keys = set()
                
            if object.update(): # Object has been interacted with. Break out.
                self.blocked = object.uid
            
            self.draw(ids, groups)
        
    def draw(self, ids, groups):
        for id in ids:
            object = self.__get_object_by_id(id)
            if not object.visible or \
                (groups and not object.group in groups):
                # Skip unwanted groups & not visible uie's
                continue
            if groups and object.group in groups: # Skip unwanted groups
                continue
            
            object.draw()
    @property
    def uids(self) -> list[int]:
        return [o.uid for o in self.QUEUE]   

UIM = UIManager()



from pig_ui.constants import *
from pig_ui.ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.ui.ui_element import UIElement

class SpecialKeyStates:
    IDLE = 0
    WAITING_FOR_DELAY = 1
    REPEATING = 2
    
class SpecialKey:
    DELAY = 0
    REPEAT = 0
    def __init__(self, delay_ms: int, repeat_ms: int):
        self.REPEAT_MS = repeat_ms
        self.DELAY_MS = delay_ms
        self.last_press_time = 0.0
        self.last_repeat_time = 0.0
        self.state = SpecialKeyStates.IDLE
        self.pressed = False
    def reset(self):
        self.state = SpecialKeyStates.IDLE
        self.pressed = False
    def update(self):
        self.pressed = False
        if self.state == SpecialKeyStates.IDLE: 
            self.state = SpecialKeyStates.WAITING_FOR_DELAY
            self.pressed = True
            self.last_press_time = time()
            self.last_repeat_time = time()
            
        elif self.state == SpecialKeyStates.WAITING_FOR_DELAY: 
            time_since = (time() - self.last_press_time) * 1000
            if time_since >= self.DELAY_MS:
                self.state = SpecialKeyStates.REPEATING
                self.last_repeat_time = time()
                
        elif self.state == SpecialKeyStates.REPEATING: 
            time_since = (time() - self.last_repeat_time) * 1000
            if time_since >= self.REPEAT_MS:
                self.pressed = True
                self.last_repeat_time = time()

class UITextInput(UIElement):
    #! Add blocking: out of bounds write
    #! Add Row/Columns
    """
    A Default TextInput
    """
    TRANSLATION_TABLE = {getattr(pg,f'K_{l}'): (l, u) for l,u in zip('0123456789abcdefghijklmnopqrstuvwxyz','=!"§$%&/()ABCDEFGHIJKLMNOPQRSTUVWXYZ')}
    TRANSLATION_TABLE[pg.K_SPACE] = (' ', ' ')
    TRANSLATION_TABLE[pg.K_RETURN] = ('', '\n')
    TRANSLATION_TABLE[pg.K_PERIOD] = ('.', ':')
    def __init__(self, app, pos, size, ux = None, draggable = False, multiline: bool = False, max_length: int = -1, type: int = 2, **kwargs):
        """
        Type 0: takes all(str)
        Type 1: takes only(int)
        Type 2: takes int & float
        """
        # click outside or returning will set is_editing to false
        """
        To reset the is_editing variable we need to inject this into UIM
        from there it will check the id & outside_clicks so it can reset
        
        in UIM:
            if self.is_editing: break # will break because the other elements should be not interacted with!
            if object.is_editing:
                if not object.hover and object.event.MOUSE_LEFT:
                    object.is_editing = False
        """
        
        # We need UXFont
        # The font can have its own anchors left, center, right
        kwargs['cb_lclick'] = self.set_edit
        kwargs['cb_unclick'] = self.reset
        if ux is None:
            
            UIELEMENT_TEXT = [
            [UXRect(-1,Color('#242424'),size=size),
                UXText(color=Color('#969696'),text_get_callback=self.get_text)],
            [UXRect(-1,Color('#242424'),size=size),
                UXText(color=Color('#ffffff'),text_get_callback=self.get_text)],
            [UXRect(-1,Color('#242424'),size=size),
                UXText(color=Color('#ffffff'),text_get_callback=self.get_text)],
            [UXRect(-1,Color('#242424'),size=size),
                UXText(color=Color('#000000'),text_get_callback=self.get_text)]
        ]
            ux = UXWrapper(UIELEMENT_TEXT)
        super().__init__(app, pos, size, ux, draggable, **kwargs)
        
        self.text = 'abc'
        self.is_editing = False
        self.pressed_keys = set()
        self.multiline = multiline
        self.max_length = max_length
        self.type = type
        self.special_keys: dict[str, SpecialKey] = {}
        
        self.set_special_key_state(pg.K_RETURN)
        self.set_special_key_state(pg.K_BACKSPACE)
        
    def get_text(self) -> str:
        return self.text
    def set_edit(self,*_):
        self.is_editing = True
    @property
    def delete(self) -> bool:
        return pg.K_BACKSPACE in self.event.KEYS
    @property
    def _return(self) -> bool:
        return pg.K_RETURN in self.event.KEYS
    @property
    def shift(self) -> bool:
        return pg.K_LSHIFT in self.event.KEYS or pg.K_RSHIFT in self.event.KEYS

    def update(self):
        return super().update()
    def set_used_keys(self):
        for key in self.event.KEYS:
            self.pressed_keys.add(key)
    
    def type_check(self, text: str) -> str:
        """
        """
        match self.type:
            case 1: 
                if text.isdecimal(): return text
                return ''
            case 2:
                try:
                    if self.text + text == '-':
                        return text
                    
                    float(self.text + text)
                    return text
                except:
                    return ''
            case 3:
                return text #! will be changed later
            case _: return text
    
    def set_special_key_state(self, key: int): 
        if key not in self.special_keys:
            self.special_keys[key] = SpecialKey(500,50)
    def update_special_key_state(self, key: int):
        if key in self.event.KEYS:
            self.special_keys[key].update()
        else:
            self.special_keys[key].reset()

    def get_special_key_state(self, key: int):
        return self.special_keys[key].pressed
    def reset(self,*_):
        self.is_editing = False
    def keyboard_interaction(self):
        if not self.is_editing: return
        self.update_special_key_state(pg.K_RETURN)
        self.update_special_key_state(pg.K_BACKSPACE)
        
        if self.get_special_key_state(pg.K_BACKSPACE):
            self.text = self.text[:-1] if self.text else ''
            return
        
        text = self.event.TEXTINPUT
        if self.max_length != -1 and len(self.text) + len(text) > self.max_length: 
            return
        text = self.type_check(text)
        if text:
            self.text += text
        if self.multiline and self.get_special_key_state(pg.K_RETURN):
            self.text += '\n'
